#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2019 cuongnv
# Licensed under the Apache License, Version 2.0, <LICENSE-APACHE or
# http://apache.org/licenses/LICENSE-2.0> or the MIT license <LICENSE-MIT or
# http://opensource.org/licenses/MIT>, at your option. This file may not be
# copied, modified, or distributed except according to those terms.


#This script can run in cronjob to period check and remove messages in rooms.

import requests
import uuid
import urllib
import psycopg2
import time
import json
import sqlite3
import os
import pickle

#Follow instruction
#1. List rooms
#2. Find admin of rooms
#3. Get access token of admin's room
#4. List old messages of that room
#5. Call redacted API clean message
#6. Clean up database

host = "<CHANGE THIS>" #change to your host
#host = "http://127.0.0.1:8008/"
#clean_room = [{timeout, [ROOM_ID], [USER]}...]
#clean_room = [(28800, ["ROOM_ID1", "ROOM_ID2".,,,], [])]

clean_room = []

db_user = "<CHANGE THIS>"
db_passwd = "<CHANGE THIS>"
db_host = "<CHANGE THIS>"
db_port = "<CHANGE THIS>"
db_name = "<CHANGE THIS>"

mutex_path = ".autoremove_v5.mutex" #To avoid run many instance at same time.
config_path = ".autoremove.conf" 	#Store master token of each room to avoid request/query to databases.

def remove_message(roomId, eventId, access_token):
    payload = {"reason": "Timeout!"}
    roomId = urllib.quote(roomId)
    eventId = urllib.quote(eventId)
    headers = {'Authorization': 'Bearer ' + access_token}
    r = requests.post(host + "_matrix/client/api/v1/rooms/"+roomId+"/redact/"+eventId , json=payload, headers=headers, verify=False)
    return r.text
#check mutex
exists = os.path.isfile(mutex_path)
if exists:
    created_time = os.path.getmtime(mutex_path)
    print "Mutex is created at: ", time.ctime(created_time)
    current_time = time.time()
    if current_time - created_time > 1000: # 1 hour ago
        print "[+] remove file because it was old!"
        os.remove(mutex_path)
    else:
        #print time.ctime(created_time)
        print("[+] other instance are running! Stop running this instance.")
        exit(0);

#first create file mutex for singleton
f = open(mutex_path, "wb")
f.close

#config files
if(os.path.isfile(config_path)):
    with open(config_path, "rb") as config_file:
        configs = pickle.load(config_file)
else:
    configs = {}

print configs

conn = psycopg2.connect(user = db_user,password = db_passwd,host = db_host,port = db_port,database = db_name)
#conn = sqlite3.connect('homeserver.db')
c = conn.cursor()
messages = {}
for config in clean_room:
    for r in config[1]:
        #get master's access_token
        sql = "SELECT token FROM access_tokens INNER JOIN rooms ON access_tokens.user_id = rooms.creator WHERE rooms.room_id ='%s'" % r
        #print(sql)
        c.execute(sql)
        token = c.fetchone()
        room_token = token[0]
        #print("Room token " + room_token)
        #get stream_ordering max by room_id
        sql = "SELECT stream_ordering FROM events WHERE room_id = '%s' AND (type='m.room.encrypted' OR type='m.room.message') ORDER BY stream_ordering DESC LIMIT 1 OFFSET 50" % r #skip 50 first rows
        c.execute(sql)
        stream_ordering = c.fetchone()
        if stream_ordering is None:
            print("[+] ignore room %s because of too few messages" % r)
            continue
        stream_ordering_max = stream_ordering[0]
        print("Stream ordering " + str(stream_ordering_max) + " of room: " + r)

        #get all message of room
        sql = "SELECT room_id, event_id, origin_server_ts, token, stream_ordering FROM events INNER JOIN access_tokens ON events.sender = access_tokens.user_id WHERE (type='m.room.encrypted' OR type='m.room.message') AND room_id = '%s'" % r

        if len(config[2]) > 0:
            user_str = ""
            for ig in config[2]:
                user_str += '\'' + ig + '\','
            user_str = user_str[:len(user_str)-1]
            sql += " AND user_id IN (%s)" % user_str

        if(r in configs):
            sql += " AND events.stream_ordering < "+ str(stream_ordering_max) +" AND events.stream_ordering > "+ str(configs[r]) + " ORDER BY events.stream_ordering ASC LIMIT 100" #get 50 row each room
        else:
            sql += " AND events.stream_ordering < "+ str(stream_ordering_max) +" ORDER BY events.stream_ordering ASC LIMIT 100" #get 50 row each room
        print ("[+] Sql: {0}".format(sql))
        c.execute(sql)
        row2 = c.fetchone()
        if row2 is not None:
            print ("[+] Delete message of room " + r)
        while row2 is not None:
            # print row2[1]
            diff = time.time() - int(row2[2])/1000
            # print ("[+] Id: {0} Diff: {1}".format(row2[1], diff))
            if diff > config[0]:
                # print ("[+] check on message with event " + row2[1])
                # print messages
                if row2[1] not in messages:
                    # print ("[+] Add messages on message")
                    messages[row2[1]] = (row2[0], row2[1], row2[3],room_token, row2[4]) #room_id, event_id, user_token, master_access_token, stream_ordering
            row2 = c.fetchone()

print ("Number of message will be removed: %d" % len(messages))

#5. Call redacted API clean message
#remove message
for key,message in messages.items():
    #check in redactions 
    redaction_sql = "select event_id from redactions where redacts = '%s'" % key
    # print redaction_sql
    c.execute(redaction_sql)
    row = c.fetchone()
    if row is not None:
        configs[message[0]] = message[4]
        print "[+] Ignore event %s because it was redacted before" % key
        continue #ignore the rest :D

    print ("[+] Remove event " + message[1])
    ret = remove_message(message[0], message[1], message[2])
    # print message
    #print(ret)
    obj = json.loads(ret)
    if "errcode" in obj:
        if obj["errcode"] == 'M_FORBIDDEN':
            print ("[+] Retry with master token")
            ret = remove_message(message[0], message[1], message[3])
            # print message
            print(ret)
        elif obj["errcode"] == 'M_LIMIT_EXCEEDED':
            time.sleep(2)
            print ("[+] Retry after timeout")
            ret = remove_message(message[0], message[1], message[2])
            # print message
            print(ret)
    else:
        configs[message[0]] = message[4]
    time.sleep(2)

#6. Clean up database (remove all event)
#Turn off remove

if messages:
    #tables = ["events", "event_json"]
    for msg in messages:
        print ("[+] Remove content of event_id %s" % msg)
        sql = 'UPDATE events SET content = \'{}\' WHERE event_id=\'%s\'' % msg
        #print sql
        c.execute(sql)
        sql = 'select json from event_json where event_id=\'%s\' limit 1' % msg
        c.execute(sql)
        row = c.fetchone()
        try:
            obj = json.loads(row[0])
            #print obj
            if "content" in obj and obj["content"] != "":
                obj["content"] = ""
                #print obj
                json_str = json.dumps(obj)
                sql = 'UPDATE event_json set json = \'%s\' where event_id=\'%s\'' % (json_str, msg)
                c.execute(sql)
                
        except:
            pass

conn.commit()
conn.close()
os.remove(mutex_path)
#save config
with open(config_path, "wb") as config_file:
    pickle.dump(configs, config_file)

